// MQTT Library
// 1.00 - 2017-07-04
// (c) Copyright 2017
//
// Zur Verwendung mit allen Allnet Libraries

#include <PubSubClient.h>
#include <Client.h>


// *** Allgemeine Globale Variablen ***

WiFiClient   MQTT_WifiClient;                                       // Der MQTT-Wifi-Client für den Verbindungsaufbau
PubSubClient MQTT_Client (MQTT_WifiClient);                         // Der MQTT-Kommunikations-Client

String MQTT_ServerURL           = "";                               // MQTT Server URL, z.B. "iot.allnet.de"
uint16 MQTT_ServerPort          = 0;                                // MQTT Server Port, normalerweise 1883
String MQTT_ServerClientID      = "";                               // Eine zufällige Unique-ID
String MQTT_ServerUserName      = "";                               // Benutzer-Name
String MQTT_ServerUserPW        = "";                               // Benutzer-Password
String MQTT_ServerTopic         = "";                               // Das Thema (als Pfad), das benutzt werden soll

String MQTT_EventQueue          = "";                               // MQTT Event-Warteschlange
bool   MQTT_EventAvail          = false;                            // Wird true, sobald ein Event verfügbar ist
uint32 MQTT_EventCurrentTime    = 0;                                // Millisekunden-Zeit des letzten Events oder 0 = Kein Event
String MQTT_EventCurrentTopic   = "";                               // Das Thema (als Pfad) des letzten Events
String MQTT_EventCurrentMessage = "";                               // Die Mitteilung zum Thema des letzten Events
bool   MQTT_EventBusy           = false;                            // True = Der Event-Handler ist grade aktiv, MQTT_EventQueue nicht verfügbar
bool   MQTT_EventTrace          = false;                            // True = Jeden Event sobald er ankommt auf dem Terminal ausgeben
bool   MQTT_ConnectTrace        = true;                             // True = Jeden Connect-Versuch auf dem Terminal ausgeben


// *** Globale Variablen für Werte vom Allnet-MQTT-Server ***

double MQTT_AllnetTemperaturHamburg      = 0.0;
double MQTT_AllnetTemperaturBerlin       = 0.0;
double MQTT_AllnetTemperaturFrankfurt    = 0.0;
double MQTT_AllnetTemperaturStuttgart    = 0.0;
double MQTT_AllnetTemperaturHannover     = 0.0;
double MQTT_AllnetTemperaturMuenchen     = 0.0;
double MQTT_AllnetTemperaturKoeln        = 0.0;
double MQTT_AllnetTemperaturDuesseldorf  = 0.0;
double MQTT_AllnetTemperaturNuernberg    = 0.0;
double MQTT_AllnetTemperaturDresden      = 0.0;
double MQTT_AllnetTemperaturNaumburg     = 0.0;
double MQTT_AllnetTemperaturHeidelberg   = 0.0;
double MQTT_AllnetTemperaturCottbus      = 0.0;
double MQTT_AllnetTemperaturSchwerin     = 0.0;
double MQTT_AllnetTemperaturMainz        = 0.0;
double MQTT_AllnetTemperaturWiesbaden    = 0.0;

double MQTT_AllnetTemperaturBremen       = 0.0;
double MQTT_AllnetTemperaturDortmund     = 0.0;
double MQTT_AllnetTemperaturKiel         = 0.0;
double MQTT_AllnetTemperaturLeipzig      = 0.0;
double MQTT_AllnetTemperaturRegensburg   = 0.0;
double MQTT_AllnetTemperaturRostock      = 0.0;
double MQTT_AllnetTemperaturSaarbruecken = 0.0;
double MQTT_AllnetTemperaturTrier        = 0.0;
double MQTT_AllnetTemperaturUlm          = 0.0;
double MQTT_AllnetTemperaturWuerzburg    = 0.0;
double MQTT_AllnetTemperaturOldenburg    = 0.0;
double MQTT_AllnetTemperaturPotsdam      = 0.0;
double MQTT_AllnetTemperaturMagdeburg    = 0.0;

double MQTT_AllnetFinanceDaxVal          = 0.0;
double MQTT_AllnetFinanceDaxDifVal       = 0.0;
double MQTT_AllnetFinanceDaxDifPerc      = 0.0;

double MQTT_AllnetFinanceMDaxVal         = 0.0;
double MQTT_AllnetFinanceMDaxDifVal      = 0.0;
double MQTT_AllnetFinanceMDaxDifPerc     = 0.0;

double MQTT_AllnetFinanceEStx50Val       = 0.0;
double MQTT_AllnetFinanceEStx50DifVal    = 0.0;
double MQTT_AllnetFinanceEStx50DifPerc   = 0.0;

double MQTT_AllnetFinanceGoldVal         = 0.0;
double MQTT_AllnetFinanceGoldDifVal      = 0.0;
double MQTT_AllnetFinanceGoldDifPerc     = 0.0;

double MQTT_AllnetFinanceOelVal          = 0.0;
double MQTT_AllnetFinanceOelDifVal       = 0.0;
double MQTT_AllnetFinanceOelDifPerc      = 0.0;

double MQTT_AllnetFinanceEURVal          = 0.0;
double MQTT_AllnetFinanceEURDifVal       = 0.0;
double MQTT_AllnetFinanceEURDifPerc      = 0.0;

double MQTT_AllnetFinanceBTCVal          = 0.0;
double MQTT_AllnetFinanceBTCDifVal       = 0.0;
double MQTT_AllnetFinanceBTCDifPerc      = 0.0;

double MQTT_AllnetFinanceETHVal          = 0.0;
double MQTT_AllnetFinanceETHDifVal       = 0.0;
double MQTT_AllnetFinanceETHDifPerc      = 0.0;


// *** MQTT-Funktionen ***

void MQTT_ClearEvents ()
{
  while (MQTT_EventBusy)    // Neuer Event wird gerade empfangen
  {
    delay(1);               // 1 Millisekunde warten
  }
  MQTT_EventQueue = "";
  MQTT_EventAvail = false;
  MQTT_EventCurrentTime = 0;
  MQTT_EventCurrentTopic = "";
  MQTT_EventCurrentMessage = "";
}

void MQTT_CallbackHandler (char* topic, byte* payload, unsigned int length)
{
  MQTT_EventBusy = true;
  uint32 m = micros();
  String Topic = cleanasc(String(topic));
  String Message = "";
  for (int i = 0; i < length; i++)
  {
    if (payload[i] != 3) // Ctrl-C ist nicht erlaubt, Event-Record-Trennzeichen
    {
      Message += (char)payload[i];
    }
  }
  if (MQTT_EventTrace)
  {
    Serial.print("Event " + String(m) + " [" + Topic + "] ");
    Serial.println(left(Message, 50));
  }
  MQTT_EventAvail = true;
  MQTT_EventQueue += String(m);
  MQTT_EventQueue += chr(9);
  MQTT_EventQueue += Topic;
  MQTT_EventQueue += chr(13);
  MQTT_EventQueue += Message;
  MQTT_EventQueue += chr(3);
  MQTT_EventBusy = false;
}

void MQTT_Disconnect ()
{
  MQTT_Client.disconnect();
}

void MQTT_Init (String serverURL, uint16 serverPort, String clientID, String userName, String userPassword)
{
  MQTT_ServerURL      = serverURL;
  MQTT_ServerPort     = serverPort;
  MQTT_ServerClientID = clientID;
  MQTT_ServerUserName = userName;
  MQTT_ServerUserPW   = userPassword;

  MQTT_Disconnect();
  MQTT_Client.setServer  (MQTT_ServerURL.c_str(), MQTT_ServerPort);
  MQTT_Client.setCallback(MQTT_CallbackHandler);
  MQTT_ClearEvents();
}

void MQTT_Init (IPAddress ipAdr, uint16 serverPort, String clientID, String userName, String userPassword)
{
  MQTT_ServerURL      = strIP(ipAdr, false);
  MQTT_ServerPort     = serverPort;
  MQTT_ServerClientID = clientID;
  MQTT_ServerUserName = userName;
  MQTT_ServerUserPW   = userPassword;

  MQTT_Disconnect();
  MQTT_Client.setServer  (ipAdr, MQTT_ServerPort);
  MQTT_Client.setCallback(MQTT_CallbackHandler);
  MQTT_ClearEvents();
}

int MQTT_GetNextEvent () // 0 = Kein Event vorhanden, 1 = Event erfolgreich gelesen, <0 = Fehlercodes
{
  if (MQTT_EventAvail)
  {
    while (MQTT_EventBusy)    // Neuer Event wird gerade empfangen
    {
      delay(1);               // 1 Millisekunde warten
    }
    int m = position(MQTT_EventQueue, chr(9), 1);
    if (m > 0)
    {
      MQTT_EventCurrentTime = val(mid(MQTT_EventQueue, 1, m - 1));
      int t = position(MQTT_EventQueue, chr(13), m + 1);
      if (t > 0)
      {
        MQTT_EventCurrentTopic = mid(MQTT_EventQueue, m + 1, t - m - 1);
        int p = position(MQTT_EventQueue, chr(3), t + 1);
        if (p > 0)
        {
          MQTT_EventCurrentMessage = mid(MQTT_EventQueue, t + 1, p - t - 1);
          MQTT_EventQueue = part(MQTT_EventQueue, p + 1);
          MQTT_EventAvail = (MQTT_EventQueue != "");
          return (1);
        }
        else
        {
          return (-3); // Error: Event-Message nicht gefunden
        }
      }
      else
      {
        return (-2); // Error: Event-Topic nicht gefunden
      }
    }
    else
    {
      return (-1); // Error: Event-Time nicht gefunden
    }
  }
  else
  {
    return (0);
  }
}

bool MQTT_Connect (String topic)
{
  if (MQTT_ConnectTrace)
  {
    Serial.print("Looking for MQTT server '" + MQTT_ServerURL + "'... ");
  }
  bool success = false;
  MQTT_Client.disconnect();
  if ((MQTT_ServerUserName == "") && (MQTT_ServerUserPW == ""))
  {
    success = MQTT_Client.connect(MQTT_ServerClientID.c_str());
  }
  else
  {
    success = MQTT_Client.connect(MQTT_ServerClientID.c_str(), MQTT_ServerUserName.c_str(),  MQTT_ServerUserPW.c_str());
  }
  if (success)
  {
    if (MQTT_ConnectTrace)
    {
      Serial.println("Connected.");
    }
    MQTT_ServerTopic = topic;
    MQTT_Client.subscribe(MQTT_ServerTopic.c_str());
   }
  else
  {
    if (MQTT_ConnectTrace)
    {
      Serial.println("");
      Serial.print("MQTT server not connected. RC = ");
      Serial.println(MQTT_Client.state());
   }
  }
  return (success);
}

bool MQTT_HandleClient ()
{
  if (MQTT_Client.connected())
  {
    MQTT_Client.loop();
    return (true);
  }
  else
  {
    bool success = MQTT_Connect(MQTT_ServerTopic);
    MQTT_Client.loop();
    return (success);
  }
}

bool MQTT_AllnetParseEventCurrent ()
{
  bool result = true;
  double v = realval(MQTT_EventCurrentMessage);
  if (MQTT_EventCurrentTopic      == "general/data/temperatur/hamburg")
  {
    MQTT_AllnetTemperaturHamburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/berlin")
  {
    MQTT_AllnetTemperaturBerlin = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/frankfurt")
  {
    MQTT_AllnetTemperaturFrankfurt = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/stuttgart")
  {
    MQTT_AllnetTemperaturStuttgart = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/hannover")
  {
    MQTT_AllnetTemperaturHannover = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/muenchen")
  {
    MQTT_AllnetTemperaturMuenchen = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/koeln")
  {
    MQTT_AllnetTemperaturKoeln = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/duesseldorf")
  {
    MQTT_AllnetTemperaturDuesseldorf = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/nuernberg")
  {
    MQTT_AllnetTemperaturNuernberg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/dresden")
  {
    MQTT_AllnetTemperaturDresden = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/naumburg")
  {
    MQTT_AllnetTemperaturNaumburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/heidelberg")
  {
    MQTT_AllnetTemperaturHeidelberg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/cottbus")
  {
    MQTT_AllnetTemperaturCottbus = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/schwerin")
  {
    MQTT_AllnetTemperaturSchwerin = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/mainz")
  {
    MQTT_AllnetTemperaturMainz = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/wiesbaden")
  {
    MQTT_AllnetTemperaturWiesbaden = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/bremen")
  {
    MQTT_AllnetTemperaturBremen = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/dortmund")
  {
    MQTT_AllnetTemperaturDortmund = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/kiel")
  {
    MQTT_AllnetTemperaturKiel = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/leipzig")
  {
    MQTT_AllnetTemperaturLeipzig = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/regensburg")
  {
    MQTT_AllnetTemperaturRegensburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/rostock")
  {
    MQTT_AllnetTemperaturRostock = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/saarbruecken")
  {
    MQTT_AllnetTemperaturSaarbruecken = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/trier")
  {
    MQTT_AllnetTemperaturTrier = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/ulm")
  {
    MQTT_AllnetTemperaturUlm = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/wuerzburg")
  {
    MQTT_AllnetTemperaturWuerzburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/oldenburg")
  {
    MQTT_AllnetTemperaturOldenburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/potsdam")
  {
    MQTT_AllnetTemperaturPotsdam = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/temperatur/magdeburg")
  {
    MQTT_AllnetTemperaturMagdeburg = v;
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/dax")    // "GER","DAX","7/4/2017","4:15pm",12468.32,"-6.99 - -0.06%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceDaxVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceDaxDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceDaxDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/mdax")   // "GER","MDAX","7/4/2017","4:15pm",24652.80,"-49.91 - -0.20%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceMDaxVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceMDaxDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceMDaxDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/estx50") // "ZRH","ESTX50 EUR P","7/4/2017","4:15pm",3490.65,"-1.16 - -0.03%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceEStx50Val = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceEStx50DifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceEStx50DifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/gold")   // "CMX","Gold Aug 17","7/4/2017","10:20am",1219.20,"-23.10 - -1.86%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceGoldVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceGoldDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceGoldDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/oel")    // "NYM","Light Sweet Crude Oil Futures,A","7/4/2017","10:20am",47.07,"+1.03 - +2.24%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceOelVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceOelDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceOelDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/eurusd")    // "CCY","EUR/USD","7/7/2017","10:15am",1.1420,"-0.0001 - -0.0131%".2615%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceEURVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceEURDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceEURDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/btc")    // "CCY","Bitcoin USD","7/7/2017","10:00am",2570.0801,"-6.9199 - -0.2685%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceBTCVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceBTCDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceBTCDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else if (MQTT_EventCurrentTopic == "general/data/finance/ether")    // "CCY","Ethereum","7/21/2017","8:29am",237.0000,"+2.2000 - +0.9370%"
  {
    int p = position(MQTT_EventCurrentMessage, ":", 1);
    if (p > 0)
    {
      p = position(MQTT_EventCurrentMessage, ",", p + 1);
      if (p > 0)
      {
        MQTT_AllnetFinanceETHVal = realval(part(MQTT_EventCurrentMessage, p + 1));
        p = position(MQTT_EventCurrentMessage, ",", p + 1);
        if (p > 0)
        {
          MQTT_AllnetFinanceETHDifVal = realval(part(MQTT_EventCurrentMessage, p + 2));
          p = position(MQTT_EventCurrentMessage, " - ", p + 1);
          if (p > 0)
          {
            MQTT_AllnetFinanceETHDifPerc = realval(part(MQTT_EventCurrentMessage, p + 3));
          }
        }
      }
    }
  }
  else
  {
    result = false;
  }
  return (result);
}
